#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# v1.0.0
# Copyright (c) 2011, Berkinet, (AKA Richard Perlman) All rights reserved.

################################################################################
from berkinet import setLogLevel, versionCheck
import time
import thread
from datetime import datetime
from datetime import timedelta
from datetime import date
import string
import re
import os
import urllib2
from urllib2 import urlopen
import socket

################################################################################
class Plugin(indigo.PluginBase):
   ########################################
	def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
		indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
		
		(self.logLevel, self.debug) = setLogLevel(self, pluginDisplayName, pluginPrefs.get("showDebugInfo1", "1"))
		
		self.irrLock = False	# Keeps the action from being called when it is running
		self.shutdown = False	# Signal flag for an orderly shutdown
		self.variableRunCheck = pluginPrefs.get("runVariableCheck", False)
		self.runFlag = True
		if self.variableRunCheck:
			flagVar = int(pluginPrefs.get("runVariable"))
			self.runFlag = indigo.variables[flagVar].value
			if self.runFlag.lower() == "false":
				self.runFlag = False
			
		if self.logLevel > 0: indigo.server.log(u"Master Irrigation Plugin Initialized", type=self.pluginDisplayName)
	
	def __del__(self):
		indigo.PluginBase.__del__(self)

	########################################
	def startup(self):
		if self.logLevel > 1: indigo.server.log(u"Plugin Starting", type=self.pluginDisplayName)
		versionCheck(self, self.pluginDisplayName, self.logLevel, self.pluginId, self.pluginVersion)
					
	def shutdown(self):
		if self.logLevel > 1: self.debugLog(u"shutdown() called")
		
	########################################
	def closedPrefsConfigUi (self, valuesDict, UserCancelled):
		if UserCancelled is False:
			if self.logLevel != int(valuesDict["showDebugInfo1"]):
				(self.logLevel, self.debug) = setLogLevel(self, self.pluginDisplayName, valuesDict["showDebugInfo1"])
				
			self.variableRunCheck = valuesDict["runVariableCheck"]
			self.runFlag = True
			if self.variableRunCheck:
				flagVar = int(valuesDict["runVariable"])
				self.runFlag = indigo.variables[flagVar].value
				if self.runFlag.lower() == "false":
					self.runFlag = False
			indigo.server.log ('Plugin preferences have been updated', type=self.pluginDisplayName)	

	######################################
	# We won't use runConcurrentThread in startup, but we will use it for an orderly shutdown.
	def stopConcurrentThread(self):
		if self.logLevel > 1: indigo.server.log(u"stopConcurrentThread called", type=self.pluginDisplayName)
		self.shutdown = True
		time.sleep(1) # Let things shut down
		if self.logLevel > 2: self.debugLog(u"Exiting stopConcurrentThread")

	########################################
	# irrrigation 
	def runSchedule(self, dailySchedDict, controllerDict, *args):
		if self.logLevel > 1: indigo.server.log(u'Starting schedule run', type=self.pluginDisplayName)
		# Now we can water
		dailySchedDict = dict(dailySchedDict)
		controllerDict = dict(controllerDict)
		for irrDev in sorted(dailySchedDict.iterkeys()):
			if self.logLevel > 2: self.debugLog(u'Starting irrigation loop. shutdown is: %s' % (self.shutdown))
			if not self.shutdown:
				if self.logLevel > 3: self.debugLog(u'Ready for zone %s:\ndict is: %s' % (irrDev, dailySchedDict[irrDev]))
				if self.logLevel > 2: self.debugLog(u'Reading schedule for zone %s' % (irrDev))
				irrTime = string.atof(dailySchedDict[irrDev]['irrTime'])
				irrZone = int(dailySchedDict[irrDev]['realOutputNumber']) -1 # Zone 1 is 0
				vDevice = dailySchedDict[irrDev]['vDevice']
				irrController = dailySchedDict[irrDev]['realIrrDevice']
				
				# First, make sure none of the controllers are running
				for controller in controllerDict:
					if self.logLevel > 2: self.debugLog(u'Checking zones on controller: %s' % (controller))
					onState = indigo.devices[int(controller)].states['activeZone']
					if self.logLevel > 2: self.debugLog(u'    Active zone: %s' % (onState))
					if onState > 0:
						if self.logLevel > 2: self.debugLog(u'    Stopping all zones on controller: %s' % (controller))
						indigo.sprinkler.stop(int(controller))
						time.sleep(5) # Give the stop command time to run
				
				# Now, start the irrigation
				if dailySchedDict[irrDev]['logCheck']:
					indigo.server.log(u'Starting irrigation for ' + irrDev + '(output ' + str(irrZone) + '), for ' + str(irrTime) + ' minutes', type=self.pluginDisplayName)
					
				indigo.sprinkler.turnOnZone(int(irrController), index=int(irrZone))
				vDevice.updateStateOnServer(key='currentlyRunning', value='on')
				vDevice.updateStateOnServer(key='lastIrrigationDuration', value=irrTime)
				now = datetime.now()					# The time we started the loop
				## for testing  ## irrTime = 1
				then = now + timedelta(minutes=irrTime)	# The time we will end the loop
				if self.logLevel > 1: indigo.server.log(u'    Starting at: %s, Expected end at: %s' % (now, then), type=self.pluginDisplayName)
				while now <= then and not self.shutdown:
					now = datetime.now()
					time.sleep(.5)
				
				indigo.sprinkler.stop(int(irrController))
				if self.logLevel > 1: indigo.server.log(u'    Actualy ended at: %s' % (now), type=self.pluginDisplayName)
				time.sleep(5) # Give the stop command time to run
				vDevice.updateStateOnServer(key='currentlyRunning', value='off')
				vDevice.updateStateOnServer(key="lastRun", value=now.strftime("%Y-%m-%d %H:%M:%S"))
				if dailySchedDict[irrDev]['logCheck']:
					indigo.server.log(u'Irrigation ended for zone ' + irrDev, type=self.pluginDisplayName)
								
		if self.logLevel > 1: indigo.server.log(u'Finished irrigation schedule run', type=self.pluginDisplayName)
		if self.shutdown:
			if self.logLevel > 0: indigo.server.log(u"Active run cancelled", type=self.pluginDisplayName)
			
		self.irrLock = False
					
	########################################
	# Action callbacks
	def createSchedule(self, action):
		
		if self.logLevel > 0: indigo.server.log(u'Starting schedule creation', type=self.pluginDisplayName)
			
		if self.irrLock:
			self.errorLog(u'Irrigation schedule already running. Action request ignored.')
			return
		else:
			self.irrLock = True
			if self.logLevel > 1: indigo.server.log(u'No lock found. Set irrLock True', type=self.pluginDisplayName)
			
		dailySchedDict = {}
		controllerDict = {}
		totalSchedTime = 0
		
		# Are we in test mode?
		testMode = False
		if str(action.pluginTypeId) == "startTestSchedule":
			if self.logLevel > 0: indigo.server.log(u'Operating in test mode', type=self.pluginDisplayName)
			testMode = True
			
		for irrDev in indigo.devices.iter('com.berkinet.masterirrigation'):
			if irrDev.enabled:
				if self.logLevel > 2 or testMode: indigo.server.log(u"Schedule create, reading device: %s" % irrDev.name, type=self.pluginDisplayName)
				tempDict = dict(irrDev.pluginProps.items() + irrDev.states.items())
				if self.logLevel > 3: self.debugLog(u"FOUND TEMPDICT: %s" % tempDict)
				# Build a list of controllers - otherwise we could just ask Indigo.
				controllerDict[tempDict['realIrrDevice']] = True
				
				# Start by figuring out if we have to water today
				irrToday = False
				if tempDict['vSchedule'] == '0':
					if self.logLevel > 2 or testMode: indigo.server.log(u"    Interval scheduling: %s days to go" % tempDict['daysToNextRun'], type=self.pluginDisplayName)
					if int(tempDict['daysToNextRun']) <= 0:
						irrToday = True
						irrTime = tempDict['vSchedPeriodTime']
						
						irrDev.updateStateOnServer(key='daysToNextRun', value=tempDict['vSchedPeriod'])
					else:
						# We won't water today, but, decrement the counter
						daysRemaining = int(tempDict['daysToNextRun']) - 1
						if not testMode: irrDev.updateStateOnServer(key='daysToNextRun', value=daysRemaining)
				elif tempDict['vSchedule'] == '1':
					today = date.weekday(datetime.today()) # Mon = 0, Sun = 6
					if today == 0 and tempDict['vSchedMon']:
						irrToday = True
						irrTime = tempDict['vSchedMonTime']
					elif today == 1 and tempDict['vSchedTue']:
						irrToday = True
						irrTime = tempDict['vSchedTueTime']
					elif today == 2 and tempDict['vSchedWed']:
						irrToday = True
						irrTime = tempDict['vSchedWedTime']
					elif today == 3 and tempDict['vSchedThu']:
						irrToday = True
						irrTime = tempDict['vSchedThuTime']
					elif today == 4 and tempDict['vSchedFri']:
						irrToday = True
						irrTime = tempDict['vSchedFriTime']
					elif today == 5 and tempDict['vSchedSat']:
						irrToday = True
						irrTime = tempDict['vSchedSatTime']
					elif today == 6 and tempDict['vSchedSun']:
						irrToday = True
						irrTime = tempDict['vSchedSunTime']
					
					if not irrToday:
						if self.logLevel > 2 or testMode: indigo.server.log(u"    Custom scheduling: not scheduled for today", type=self.pluginDisplayName)
					else:
						if self.logLevel > 2 or testMode: indigo.server.log(u"    Scheduled for today", type=self.pluginDisplayName)
						
				# Now, add the information for our schedule to a dictionary...
				if irrToday:
					if tempDict['multiplyVariableCheck']:
						theVar = indigo.variables[int(tempDict['multiplyVariable'])]
						origIrrTime = irrTime
						irrTime = float(irrTime)
						# We multiply, round and then convert to a string to pass the value correctly
						irrTime = str(round(irrTime * float(theVar.value), 2))
						if self.logLevel > 2: indigo.server.log(u"    Used multiply var:%s , changed orig time: %s to %s" % (str(theVar.value), origIrrTime, irrTime))
					else:
						irrTime = str(float(irrTime))
					
					realIrrDevice = int(tempDict['realIrrDevice'])
					realOutputNumber = int(tempDict['realOutputNumber'])
					
					if float(irrTime) > 0:
						dailySchedDict[irrDev.name] = {'vDevice':irrDev, 'realIrrDevice':realIrrDevice, 'realOutputNumber':realOutputNumber, 'irrTime':irrTime, 'logCheck':tempDict['logCheck']}
						totalSchedTime += float(irrTime)
						if self.logLevel > 2: indigo.server.log(u"    Adding device: %s to schedule" % irrDev.name)
					else:
						if self.logLevel > 2: indigo.server.log(u"    Run time = 0: %s, not added to schedule" % irrDev.name)
					
		# Now it is time to start watering, unless we are in test mode		
		if self.logLevel > 3: self.debugLog(u'Finished irrigation setup: dict is: %s' % dailySchedDict)
	
		schedReport = ""
		for zone in sorted(dailySchedDict.iterkeys()):
			schedReport += zone + ": " + dailySchedDict[zone]['irrTime'] + "; "
		if totalSchedTime > 0:
			if self.logLevel > 0 or testMode: indigo.server.log(u"Total scheduled irrigation time is: " + str(totalSchedTime) + " mins, for the following zones:\n                                    " + schedReport, type=self.pluginDisplayName)
		else:
			if self.logLevel > 0 or testMode: indigo.server.log(u'Finished schedule creation: Nothing scheduled.', type=self.pluginDisplayName)
		
		if self.logLevel > 3: self.debugLog('Runcheck: %s, Runcheck value: %s' % (self.variableRunCheck, self.runFlag))		
		if self.variableRunCheck and not testMode:
			if not self.runFlag:
				if self.logLevel > 0: indigo.server.log(u'Schedule cancelled. Indigo run control variable was False', type=self.pluginDisplayName)
				self.irrLock = False
				return
		
		if not testMode and len(dailySchedDict) > 0:
			self.shutdown = False	
			thread.start_new_thread(self.runSchedule, (dailySchedDict, controllerDict, 1))
		else:
			if self.logLevel > 1: indigo.server.log(u'Set irrLock False', type=self.pluginDisplayName)
			self.irrLock = False
			if testMode:
				if self.logLevel > 0: indigo.server.log(u'Ending test mode', type=self.pluginDisplayName)

	def stopSchedule(self, action):
		indigo.server.log(u'Stopping schedule', type=self.pluginDisplayName)
		if self.irrLock:
			if self.logLevel > 0: indigo.server.log(u'Irrigation schedule was running. Now stopped.', type=self.pluginDisplayName)
			self.shutdown = True
		else:
			if self.logLevel > 0: indigo.server.log(u'Irrigation schedule was not running.', type=self.pluginDisplayName)
	
	def resetSchedule(self, action):
		if self.logLevel > 0: indigo.server.log(u'Resetting schedule', type=self.pluginDisplayName)
		for irrDev in indigo.devices.iter('com.berkinet.masterirrigation'):
			irrDev.updateStateOnServer(key='daysToNextRun', value='0')

	def singleZoneSchedule(self, action):
		if self.logLevel > 1: indigo.server.log(u'Single Zone schedule', type=self.pluginDisplayName)
		if self.irrLock:
			self.errorLog(u'Irrigation schedule already running. Action request ignored.')
			return
		else:
			self.irrLock = True
			if self.logLevel > 1: indigo.server.log(u'No lock found. Set irrLock True', type=self.pluginDisplayName)
			
		irrDev = indigo.devices[action.deviceId]
		irrTime = float(action.props['runTime'])
		realIrrDevice = irrDev.pluginProps['realIrrDevice']
		realOutputNumber = irrDev.pluginProps['realOutputNumber']
		logCheck = irrDev.pluginProps['logCheck']
		
		if self.logLevel > 2: self.debugLog(u'Found: %s, runTime=%s, device=%s, output=%s, logCheck=%s' % (irrDev.name, irrTime, realIrrDevice, realOutputNumber, logCheck))
		dailySchedDict = {}
		dailySchedDict[irrDev.name] = {'vDevice':irrDev, 'realIrrDevice':realIrrDevice, 'realOutputNumber':realOutputNumber, 'irrTime':irrTime, 'logCheck':logCheck}
		if self.logLevel > 2: self.debugLog(u"    Adding device: %s to schedule" % irrDev.name)
		controllerDict = {}
		controllerDict[realIrrDevice] = True
		if self.logLevel > 1: indigo.server.log(u'Finished single run schedule creation', type=self.pluginDisplayName)	
		if self.logLevel > 2: 
			if self.debug:
				schedReport = irrDev.name + ": " + str(irrTime) + "; "
				self.debugLog(u"    The schedule is: %s" % (schedReport))
		
		thread.start_new_thread(self.runSchedule, (dailySchedDict, controllerDict, 1))